"""

Copyright (c) 2019 The Cereja Project

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import os
import time
import unittest
import logging

from cereja.array import group_items_in_batches, remove_duplicate_items, flatten, \
    array_gen, get_shape, Matrix
from cereja.utils import is_iterable, is_sequence, chunk
from cereja.array import prod
from cereja.mathtools import theta_angle
from cereja.config.cj_types import Number
from cereja.mltools import Corpus
from cereja.mltools.pln import LanguageData
from cereja.display import State, Progress
from cereja.system import Path, TempDir, mkdir
from cereja.system.unicode import Unicode
from cereja.utils import CjTest
from cereja.hashtools import base64_decode, base64_encode, is_base64, md5

logger = logging.getLogger(__name__)


class UtilsTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.shapes = (2, 2, 3), (1, 3, 2, 1, 1), (2, 1, 1)
        self.sequences = [
            [[[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]], [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]]],

            [[[[[0.0]], [[0.0]]], [[[0.0]], [[0.0]]], [[[0.0]], [[0.0]]]]],

            [[[0.]], [[0.]]]
        ]

    def test_is_iterable(self):
        self.assertFalse(is_iterable(1))
        self.assertTrue(is_iterable('hi'))
        self.assertTrue(is_iterable([]))

    def test_theta_angle(self):
        u = (2, 2)
        v = (0, -2)
        self.assertEqual(theta_angle(u, v), 135.0)

    def test_remove_duplicate_items(self):
        test_values_sample_list = [([1, 2, 1, 2, 1], [1, 2]),
                                   (['hi', 'hi', 'ih'], ['hi', 'ih']),
                                   ]
        test_values_list_of_list = [([['hi'], ['hi'], ['ih']], [['hi'], ['ih']]),
                                    ([[1, 2, 1, 2, 1], [1, 2, 1, 2, 1], [1, 2, 1, 2, 3]],
                                     [[1, 2, 1, 2, 1], [1, 2, 1, 2, 3]])
                                    ]

        for test_value, expected_value in test_values_sample_list:
            self.assertEqual(remove_duplicate_items(test_value), expected_value)

        for test_value, expected_value in test_values_list_of_list:
            self.assertEqual(remove_duplicate_items(test_value), expected_value)

    def test_flatten(self):
        sequence = [[1, 2, 3], [], [[2, [3], 4], 6]]
        expected_value = [1, 2, 3, 2, 3, 4, 6]
        self.assertEqual(flatten(sequence), expected_value)

        max_recursion = 2
        expected_value = [1, 2, 3, 2, [3], 4, 6]
        self.assertEqual(flatten(sequence, max_recursion=max_recursion), expected_value)

    def test_is_sequence(self):
        sequence = [[1, 2, 3], [], [[2, [3], 4], 6]]
        self.assertTrue(is_sequence(sequence))

        self.assertFalse(is_sequence('hi cereja'))
        self.assertTrue(is_sequence(dict({'k1': 1, 'k2': 2}).keys()))

    def test_array_gen(self):
        for args, expected in zip(self.shapes, self.sequences):
            data = array_gen(args)
            self.assertEqual(data, expected)

        with_kwargs = [({"shape": (2, 1, 1), "v": ['hail', 'pythonics!']}, [[['hail']], [['pythonics!']]])
                       ]

        for kwargs, expecteds in with_kwargs:
            result = array_gen(**kwargs)
            self.assertEqual(result, expecteds)

    def test_get_shape(self):
        for expected_shape, seq in zip(self.shapes, self.sequences):
            shape_received = get_shape(seq)
            sequence_length = len(flatten(seq))
            self.assertEqual(sequence_length, prod(shape_received), msg=f"array shape {shape_received} is inconsistent")
            self.assertEqual(shape_received, expected_shape)

    def test_randn(self):
        logger.warning("Awaiting tests!")

    def test_matrix(self):
        a = Matrix([[1, 2, 3], [1, 2, 3]])
        b = Matrix([[1, 2, 3], [1, 2, 3]])
        expected = Matrix([[1.0, 1.0, 1.0],
                           [1.0, 1.0, 1.0]])
        self.assertTrue((a / b) == expected)


class PathTest(unittest.TestCase):
    def test_sanity(self):
        self.assertEqual(Path(__file__).parent, Path(os.path.realpath(__file__)).parent)
        p_test = 'cereja/test/sanity'
        p = Path(p_test)
        self.assertTrue(p.name, 'sanity')
        self.assertTrue(p.parent.name, 'test')
        self.assertTrue(p == p_test)
        self.assertTrue('sanity' in p)
        p = p + ['con', 'cat']
        p_test = Path('cereja/test/sanity').join('con', 'cat')
        self.assertEqual(p_test.parts[-2:], ('con', 'cat'))
        self.assertTrue(p == p_test)
        self.assertListEqual(Path(__file__).parent.list_dir(only_name=True),
                             list(map(lambda x: x.rsplit('.')[0], os.listdir(Path(__file__).parent))))

        with TempDir() as tmp_dir:
            tmp_dir = Path(tmp_dir)
            suffix_case = tmp_dir.join('test.suffix')
            self.assertEqual(suffix_case.suffix, '.suffix')
            mkdir(suffix_case)
            self.assertEqual(suffix_case.suffix, '')


class UnicodeToolTestCase(unittest.TestCase):
    def test_sanity(self):
        unicode_ = 0x1F352
        # instance
        instance_ = Unicode.parse(unicode_)
        # expected
        test_ = CjTest(instance_)
        test_.add_check(test_.name == "CHERRIES",
                        test_.value == '\U0001F352',
                        test_.decimal == 127826,
                        test_.bin == "0b11111001101010010",
                        test_.hex == "0x1f352",
                        test_.oct == '0o371522'
                        )

        test_.check_all()


class CorpusTestCase(unittest.TestCase):
    def test_sanity(self):
        source = ['how are you?', 'my name is Joab', 'I like coffee', 'how are you joab?', 'how', 'we are the world']
        target = ['como você está?', 'meu nome é Joab', 'Eu gosto de café', 'Como você está joab?', 'como',
                  'Nós somos o mundo']

        corpus = Corpus(source_data=source, target_data=target, source_name='en', target_name='pt')
        self.assertEqual(str(corpus), 'Corpus(examples: 6 - source_vocab_size: 13 - target_vocab_size:15)')
        self.assertEqual(str(corpus.source), 'LanguageData(examples: 6 - vocab_size: 13)')
        self.assertEqual(str(corpus.target), 'LanguageData(examples: 6 - vocab_size: 15)')

        self.assertDictEqual(dict(corpus.source.phrases_freq.items()),
                             {'how are you': 1, 'my name is joab': 1, 'i like coffee': 1, 'how are you joab': 1,
                              'how':         1, 'we are the world': 1})

        self.assertDictEqual(dict(corpus.source.words_freq.items()),
                             {'how':  3, 'are': 3, 'you': 2, 'joab': 2, 'my': 1, 'name': 1, 'is': 1, 'i': 1,
                              'like': 1, 'coffee': 1, 'we': 1, 'the': 1, 'world': 1})
        self.assertDictEqual(dict(corpus.target.phrases_freq.items()),
                             {'como você está':      1, 'meu nome é joab': 1, 'eu gosto de café': 1,
                              'como você está joab': 1, 'como': 1, 'nós somos o mundo': 1})
        self.assertDictEqual(dict(corpus.target.words_freq.items()),
                             {'como':  3, 'você': 2, 'está': 2, 'joab': 2, 'meu': 1, 'nome': 1, 'é': 1, 'eu': 1,
                              'gosto': 1, 'de': 1, 'café': 1, 'nós': 1, 'somos': 1, 'o': 1, 'mundo': 1})


class ProgressTestCase:
    def test_sanity(self):
        progress = Progress("Cereja Progress Test")
        progress_test = CjTest(progress)

        class MyCustomSate(State):

            def display(self, current_value: Number, max_value: Number, current_percent: Number, time_it: Number,
                        n_times: int) -> str:
                return "RUNNING"

            def done(self, current_value: Number, max_value: Number, current_percent: Number, time_it: Number,
                     n_times: int) -> str:
                return "FINISHED"

        state = MyCustomSate()
        progress.add_state(state)

        progress_test.add_check(
                progress_test.name == "Cereja Progress Test",
                progress_test._awaiting_update == True,
                progress_test._show == False,
                progress_test.started_time == None,
                progress_test._err == False,
                progress_test._was_done == False,
                progress_test._with_context == False
        )
        progress_test.check_all()

        with Progress("p") as p:
            test_p = CjTest(p)
            test_p.add_check(test_p.name == 'p',
                             test_p._with_context == True,
                             test_p.started == True,
                             test_p._awaiting_update == True,
                             test_p.max_value == 100
                             )
            test_p.check_all()
            time.sleep(3)
            p_iterator = p(range(1, 500))
            test_p.add_check(test_p.max_value == 499)
            test_p.check_all()
            for i in p_iterator:
                time.sleep(1 / i)
                if i == 1:
                    test_p.add_check(
                            test_p._awaiting_update == False,
                            test_p.max_value == 499,
                            test_p._current_value == 2
                    )
                    test_p.check_all()

            for i in p(range(1, 500)):
                time.sleep(1 / i)

            for i in p(range(1, 500)):
                time.sleep(1 / i)


class LanguageDataTestCase(unittest.TestCase):
    def test_sanity(self):
        data = LanguageData(['how are you?', 'I like it', 'cereja', 'cherry'], to_lower=True)
        self.assertEqual(data.synergy(['how are you?', 'I like it', 'cereja', 'cherry']), 1)
        self.assertEqual(data.synergy(['how are you?', 'I like it', 'cereja']), 0.889)


class HashToolsTestCase(unittest.TestCase):
    def test_sanity(self):
        vals = ([1, 2, '3'],
                (1, '2', '1'),
                {'hi': 1, 'hello': 3, 'deep': {'oi'}},
                1,
                2,
                'a')
        for val in vals:
            res = base64_encode(val)
            self.assertTrue(is_base64(res))
            self.assertEqual(val, base64_decode(res, eval_str=True))

        self.assertEqual('fc4d5dcfe7146d09b7fbad980fb90797', md5('set the alarm clock for seven of the clock .'))


if __name__ == '__main__':
    unittest.main()



import copy
import csv
import logging
import os
import re
import pickle
import tempfile
from abc import ABCMeta, abstractmethod, ABC
from collections import ValuesView
from io import BytesIO
from typing import Any, List, Union, Type, TextIO, Iterable, Tuple, KeysView, ItemsView
import json
from .._requests import request
from zipfile import ZipFile, ZIP_DEFLATED

from cereja.system import Path, mkdir
from cereja.array import get_cols, flatten, get_shape
from cereja.utils import is_sequence, clipboard
from cereja.system import memory_of_this
from cereja.utils import string_to_literal, sample, fill

logger = logging.Logger(__name__)

__all__ = ['FileIO']


class _IFileIO(metaclass=ABCMeta):
    _size_map = {"B":  1.e0,
                 "KB": 1.e3,
                 "MB": 1.e6,
                 "GB": 1.e9,
                 "TB": 1.e12
                 }
    _dont_read = [".pyc"]
    _ignore_dir = [".git"]
    _date_format = "%Y-%m-%d %H:%M:%S"
    # need implement
    _is_byte = None
    _only_read = None
    _newline = True
    _ext_allowed = None
    _need_implement = {'_is_byte':     """
    Define the configuration of the file in the class {class_name}:
    e.g:
    class {class_name}:
        _is_byte: bool = True # or False
    """,
                       '_only_read':   """Define the configuration of the file in the class {class_name}:
    e.g:
    class {class_name}:
        _only_read: bool = True # or False

    """,
                       '_ext_allowed': """Define the configuration of the file in the class {class_name}:
                       e.g:
                       class {class_name}:
                           _ext_allowed: bool = '.txt' # or ''

                       """
                       }

    def __init__(self):
        self._built = False
        self._length = None
        self._last_update = None
        self._is_deleted = False
        self._current_change = 0
        self._max_history_length = 50
        self._change_history = []
        self._data = None

    @property
    @abstractmethod
    def data(self) -> Any:
        pass

    @property
    @abstractmethod
    def path(self) -> Path:
        pass

    @property
    @abstractmethod
    def name(self):
        pass

    @property
    @abstractmethod
    def ext(self):
        pass

    @property
    @abstractmethod
    def name_without_ext(self):
        pass

    @property
    @abstractmethod
    def dir_name(self):
        pass

    @property
    @abstractmethod
    def dir_path(self):
        pass

    @property
    @abstractmethod
    def is_empty(self):
        pass

    @property
    @abstractmethod
    def updated_at(self):
        pass

    @property
    @abstractmethod
    def created_at(self):
        pass

    @property
    @abstractmethod
    def was_changed(self) -> bool:
        pass

    @property
    @abstractmethod
    def length(self):
        pass

    @property
    @abstractmethod
    def history(self):
        pass

    @property
    @abstractmethod
    def exists(self):
        pass

    @property
    @abstractmethod
    def last_access(self):
        pass

    @abstractmethod
    def _load(self, mode=None, **kwargs) -> Any:
        pass

    @abstractmethod
    def save(self, on_new_path: Union[str, Path] = None, exist_ok=False, overwrite=False, **kwargs):
        pass

    @abstractmethod
    def _parse_fp(self, fp: Union[TextIO, BytesIO]) -> Any:
        pass

    @abstractmethod
    def _save_fp(self, fp: Union[TextIO, BytesIO]) -> None:
        pass

    @abstractmethod
    def undo(self):
        pass

    @abstractmethod
    def redo(self):
        pass

    @abstractmethod
    def parse(self, data):
        pass

    @abstractmethod
    def add(self, data, **kwargs):
        pass

    @abstractmethod
    def size(self, unit: str = "KB"):
        pass

    @abstractmethod
    def delete(self):
        pass

    @abstractmethod
    def __iter__(self):
        pass

    @abstractmethod
    def set_path(self, p: str):
        pass


class _FileIO(_IFileIO, ABC):

    def __init__(self, path_: Path, data=None, creation_mode=False, is_byte=None, **kwargs):
        super().__init__()
        if is_byte is not None:
            self._is_byte = bool(is_byte)
        if not isinstance(path_, Path):
            path_ = Path(path_)
        self._creation_mode = creation_mode
        self._path = path_
        if self._ext_allowed:
            msg_assert_err = f'{path_.suffix} != {self._ext_allowed}. Force with support_ext = ".xyz" '
            assert self._path.suffix in self._ext_allowed, msg_assert_err
        self._ext_without_point = self._path.suffix.strip(".").upper()
        self._built = True  # next setattr will be saved in history to be future recovery
        if creation_mode:
            self._data = self.parse(data)
        else:
            self._data = self._load(**kwargs)

    def __init_subclass__(cls, **kwargs):
        for attr in cls._need_implement:
            if getattr(cls, attr) is None:
                raise NotImplementedError(cls._need_implement.get(attr).format(class_name=cls.__name__))

    def __repr__(self):
        return f"File{self.__class__.__name__.replace('_', '')}({self._path.name})"

    def __setattr__(self, key, value):
        object.__setattr__(self, key, copy.copy(value))
        if hasattr(self, '_change_history') and key not in (
                '_current_change', '_max_history_length', '_change_history', '_built',
                '_last_update') and self._built is True:
            self._set_change(key, copy.copy(object.__getattribute__(self, key)))  # append last_value of attr

    def __getitem__(self, item):
        try:
            return self.data[item]
        except TypeError:
            raise TypeError(f"'{self.__class__.__name__}' object is not subscriptable")

    def __setitem__(self, key, value):
        # TODO: save only the index that has been changed
        try:
            self._data[key] = value
            self._set_change('_data', copy.copy(self._data))
        except TypeError:
            raise TypeError(f"'{self.__class__.__name__}' object is not subscriptable")

    def _set_change(self, key, value):
        self._change_history = self._change_history[:self._current_change + 1]
        if len(self._change_history) >= self._max_history_length:
            self._change_history.pop(0)
        self._change_history.append((key, value))
        self._current_change = len(self._change_history)

    def _select_change(self, index):
        try:

            key, value = self._change_history[self._current_change + index]
            object.__setattr__(self, key, copy.copy(value))
            self._current_change += index
            logger.warning(f'You selected amendment {self._current_change + 1}')
        except IndexError:
            logger.info("It's not possible")

    @property
    def line_sep(self):
        return os.linesep

    @property
    def was_changed(self) -> bool:
        """
        Detect if file is changed by another application
        if file not exists return False

        @return: bool
        """
        if not self.exists:
            return False
        return self._last_update is not None and self.updated_at != self._last_update

    @property
    def length(self):
        return len(self._data)

    @property
    def history(self):
        return self._change_history

    @property
    def updated_at(self):
        return self._path.updated_at

    @property
    def created_at(self):
        return self._path.created_at

    @property
    def last_access(self):
        return self._path.last_access

    @property
    def name(self):
        return self._path.name

    @property
    def ext(self):
        return self._path.suffix

    @property
    def name_without_ext(self):
        return self._path.stem

    @property
    def data(self):
        return copy.copy(self._data)

    @property
    def path(self):
        return self._path

    @property
    def dir_name(self):
        return self._path.parent_name

    @property
    def dir_path(self):
        return self._path.parent

    @property
    def is_empty(self):
        return not bool(self._data)

    @property
    def exists(self):
        return self.path.exists

    @property
    def only_read(self) -> bool:
        return self._only_read

    @classmethod
    def add(cls, *args, **kwargs):
        raise NotImplementedError("it's not implemented")

    @classmethod
    def parse_ext(cls, ext: str) -> str:
        if not isinstance(ext, str):
            raise TypeError('Error during parsing ext. Send a string.')
        return '.' + ext.replace('.', '')

    def _get_mode(self, mode) -> str:
        """
        if mode is 'r' return 'r+' or 'r+b' else mode is 'w' return 'w+' or 'w+b'
        """
        return f'{mode}+b' if self._is_byte else f'{mode}+'

    def size(self, unit: str = "KB"):
        """
        returns the size that the file occupies on the disk.

        :param unit: choose anyone in ('B', 'KB', 'MB', 'GB', 'TB')

        """
        assert isinstance(unit, str), f"expected {str.__name__} not {type(unit).__name__}."

        unit = unit.upper()

        assert unit in self._size_map, f"{repr(unit)} is'nt valid. Choose anyone in {tuple(self._size_map)}"
        # subtracts the size of the python object
        return (memory_of_this(self._data) - self._data.__class__().__sizeof__()) / self._size_map[unit]

    def _load(self, mode=None, **kwargs) -> Any:
        if self._path.suffix in self._dont_read:
            logger.warning(f"I can't read this file. See class attribute <{self.__class__.__name__}._dont_read>")
            return
        mode = mode or self._get_mode('r')
        encoding = None if self._is_byte else 'utf-8'
        assert 'r' in mode, f"{mode} for read isn't valid."
        try:
            with open(self._path.path, mode=mode, encoding=encoding, **kwargs) as fp:
                parsed = self._parse_fp(fp)
                self._last_update = self.updated_at
                return parsed
        except PermissionError as err:
            logger.error(err)
            return
        except Exception as err:
            raise IOError(f"Error when trying to read the file {self._path}\n{err}")

    def save(self, on_new_path: Union[str, Path] = None, mode=None, exist_ok=False, overwrite=False, force=False,
             encoding=None,
             **kwargs):

        if not force:
            assert self._only_read is False, f"{self.name} is only read."

            if on_new_path is not None:
                self.set_path(on_new_path)

            if self.was_changed:
                assert overwrite is True, AssertionError(
                        f"File change detected (last change {self.updated_at}), if you want to "
                        + f"overwrite set overwrite=True")

            assert (exist_ok or not self.path.exists) or overwrite is True or force is True, FileExistsError(
                    "File exists. If you want override, please send 'exist_ok=True'")

        mode = mode or self._get_mode('w')
        assert 'w' in mode, f"{mode} for write isn't valid."
        encoding = None if self._is_byte else 'utf-8'
        with tempfile.TemporaryDirectory() as tmpdirname:
            tmp_file = Path(tmpdirname).join(self.name)
            if not self._newline and not self._is_byte:
                kwargs.update({'newline': ''})
            try:
                with open(tmp_file, mode=mode, encoding=encoding, **kwargs) as fp:
                    self._save_fp(fp)
            except Exception as err:
                raise IOError(f'Error writing to the file system file: {err}')
            tmp_file.cp(self.path)
            self._last_update = self.updated_at

    def delete(self):
        self._path.rm()
        self._is_deleted = True

    def set_path(self, path_):
        p = Path(path_)
        assert not p.is_dir, f"{p.path} is a directory."
        self._path = p

    def undo(self):
        if self._current_change > 0:
            index = -2 if self._current_change == len(self._change_history) else -1
            self._select_change(index)

    def redo(self):
        if self._current_change < len(self._change_history):
            self._select_change(+1)

    def sample_items(self, k: int = 1, is_random=False):
        """
        Get sample of this.

        @param k: length of sample
        @param is_random: bool
        """
        assert isinstance(k, int) and k > 0, f"k = {k} isn't valid."
        return sample(self.data, k=k, is_random=is_random)

    def __len__(self):
        if self._length is None:
            return len(self._data)
        return self._length

    def __next__(self):
        for i in self._data:
            yield i

    def __iter__(self):
        return self.__next__()

    @classmethod
    def load(cls, path_, **kwargs):
        return cls(path_, creation_mode=False, **kwargs)

    @classmethod
    def create(cls, path_, data=None, **kwargs):
        return cls(path_=path_, data=data, creation_mode=True, **kwargs)


class _Generic(_FileIO):
    _is_byte: bool = True
    _only_read = False
    _ext_allowed = ()
    _split_lines = False

    def _parse_fp(self, fp: BytesIO) -> List[bytes]:
        if self._split_lines or not bool(self.ext):
            data = fp.read().splitlines()
        else:
            data = fp.read()
        return self.parse(data)

    def _save_fp(self, fp: Union[TextIO, BytesIO]) -> None:
        fp.write(self._convert('\n').join(i if isinstance(i, (str, bytes)) else str(i) for i in self._data))

    def _convert(self, val):
        if self._is_byte and not isinstance(val, bytes):
            return str(val).encode()
        if not self._is_byte and isinstance(val, str):
            return string_to_literal(val)
        return val

    def parse(self, data) -> List[bytes]:
        if isinstance(data, (bytes, str, int, float, complex)):
            return [self._convert(data)]
        if not data:
            return [self._convert('')]
        if isinstance(data, (list, tuple)):
            return [self._convert(i) for i in data]
        raise ValueError(f"{type(data)} isn't valid.")

    def add(self, data, line=-1):
        data = self.parse(data)
        if line != -1:
            for pos, i in enumerate(data, line):
                self._data.insert(pos, i)
        else:
            self._data += data
            return
        self._set_change('_data', self._data.copy())

    def remove(self, line: Union[int, str]):
        self._data.pop(line)
        self._set_change('_data', self._data.copy())


class _TxtIO(_Generic):
    _is_byte: bool = False
    _only_read = False
    _ext_allowed = ()
    _split_lines = True


class _JsonIO(_FileIO):
    _is_byte: bool = False
    _only_read = False
    _ext_allowed = ('.json',)
    indent = False
    ensure_ascii = False

    def __init__(self, path_: Path, indent=False, **kwargs):
        self.indent = indent
        if 'ensure_ascii' in kwargs:
            self.ensure_ascii = kwargs.pop('ensure_ascii')
        super().__init__(path_, **kwargs)

    def _parse_fp(self, fp: TextIO) -> dict:
        return json.load(fp)

    def _save_fp(self, fp):
        json.dump(self._data, fp, indent=4 if self.indent else None, ensure_ascii=self.ensure_ascii)

    def items(self) -> ItemsView:
        return self._data.items()

    def keys(self) -> KeysView:
        return self._data.keys()

    def values(self) -> ValuesView:
        return self._data.values()

    def parse(self, data) -> dict:
        if isinstance(data, dict):
            return data
        elif isinstance(data, (str, bytes)):
            return json.loads(data)
        else:
            raise TypeError(f"{type(data)} isn't valid.")

    def _add(self, k, v):
        self._data[k] = v

    def add(self, key, value):
        self._add(key, value)

    def remove(self, key):
        self._data.pop(key)

    def get(self, _key):
        return self._data.get(_key)

    def __getitem__(self, item):
        try:
            return self.data[item]
        except KeyError:
            raise KeyError(f"{item} not found.")

    def __setitem__(self, key, value):
        self.add(key, value)

    def __iter__(self):
        for i in self.data:
            yield i


class _Mp4IO(_FileIO, ABC):
    _is_byte: bool = True
    _only_read = True
    _ext_allowed = ('.mp4',)

    def _save_fp(self, fp):
        fp.write(self._data)

    def _parse_fp(self, fp: BytesIO) -> bytes:
        return fp.read()

    def parse(self, data):
        return self._data


class _CsvIO(_FileIO):
    _is_byte: bool = False
    _only_read = False
    _newline = False
    _ext_allowed = ('.csv',)

    def __init__(self, *args, cols: Union[Tuple[str], List[str]] = (), fill_with=None,
                 str_to_literal=True, has_col=False, **kwargs):
        """
        :param data: expected row or list of rows with `n` elements equal to `k` cols, if not it will be fill with
        `fill_with` value.
        :param fill_with: fill empty
        :param has_col: bool
        :param literal_values: Convert `str` on literal types e.g `"'1'"` (str) -> `1` (int)
        """
        self._cols = cols
        self._str_to_literal = str_to_literal
        self._fill_with = fill_with
        self._n_values = 0
        self._has_col = has_col
        super().__init__(*args, **kwargs)

    @property
    def cols(self):
        return tuple(value if value is not None else f'unamed_col_{i}' for i, value in
                     enumerate(self._cols))

    @property
    def n_cols(self):
        return len(self.cols)

    @property
    def n_values(self):
        return self._n_values

    @property
    def rows(self):
        """
        generator for each row

        format -> {col_name: value, col_name2: value, ...}

        :return: rows with column. Is a dict type
        """
        for row in self.data:
            yield dict(zip(self.cols, row))

    def _parse_fp(self, fp: TextIO, **kwargs) -> List[List[Any]]:
        reader = csv.reader(fp)
        return self.parse(reader)

    def _save_fp(self, fp: TextIO):
        writer = csv.DictWriter(fp, fieldnames=self.cols)
        writer.writeheader()
        writer.writerows(self.rows)

    def _normalize_row(self, row, fill_with=None, literal_values=True):
        if not is_sequence(row):
            row = [row]
        if self._cols:
            row = fill(value=list(row), max_size=len(self.cols), with_=fill_with) if len(row) < len(self.cols) else row
        return [string_to_literal(item) if isinstance(item, str) else item for item in row if literal_values]

    def parse(self, data: Union[Iterable[Iterable[Any]], str, int, float, complex], fill_with=None) -> List[List[Any]]:
        if fill_with is None:
            fill_with = self._fill_with
        normalized_data = []
        if isinstance(data, dict):
            self._cols = tuple(data)
            data = data.values()
        if not is_sequence(data):
            data = [data]
        for row in data:
            if not self._cols:
                # set cols on first iter
                self._cols = fill(value=list(row), max_size=len(row), with_=fill_with)
                continue
            row_normalized = self._normalize_row(row, fill_with=fill_with, literal_values=self._str_to_literal)
            self._n_values += len(row_normalized)
            normalized_data.append(row_normalized)
        return normalized_data

    def to_dict(self):
        return dict(zip(self.cols, get_cols(self.data)))

    def flatten(self):
        return flatten(self._data)

    def shape(self):
        return get_shape(self._data)

    def _add(self, row: List[Any], index, fill_with=None):
        row = self.parse([row], fill_with=fill_with)
        if index != -1:
            for pos, i in enumerate(row, index):
                self._data.insert(pos, i)
        else:
            self._data += row

    def add(self, row: List[Any], index=-1, fill_with=None):
        """
        Add row is similar to list.append

        :param row: expected row with n elements equal to k cols, if not it will be fill with `fill_with` value.
        :param index: position that will be added
        :param fill_with: None
        """
        self._add(row=row, index=index, fill_with=fill_with)

    def __getitem__(self, item):
        if isinstance(item, str):
            if item not in self.cols:
                raise ValueError(f'Column name <{item}> not in {self.cols}')
            return get_cols(self.data)[self.cols.index(item)]
        if isinstance(item, tuple):
            lines, col = item
            if isinstance(col, tuple):
                raise ValueError("Isn't Possible.")
            assert col <= self.n_cols - 1, ValueError(
                    f"Invalid Column. Choice available index {list(range(self.n_cols))}")
            return get_cols(self.data)[col][lines]
        return super().__getitem__(item)


class _PyObj(_FileIO):
    _is_byte: bool = True
    _only_read = False
    _newline = False
    _ext_allowed = ('.pkl',)

    def _parse_fp(self, fp: TextIO, **kwargs) -> List[List[Any]]:
        loaded_obj = pickle.load(fp, **kwargs)
        return self.parse(loaded_obj)

    def _save_fp(self, fp: TextIO):
        pickle.dump(self.data, fp)

    def parse(self, data):
        return data


class _ZipFileIO(_FileIO):
    _is_byte: bool = False
    _only_read = False
    _newline = False
    _ext_allowed = ('.zip',)

    def __init__(self, path_: Path, **kwargs):

        super().__init__(path_, **kwargs)

    def add(self, data: Union[str, list, tuple]):
        parsed = self.parse(data)
        self._data += parsed

    def _parse_fp(self, fp: Union[TextIO, BytesIO]) -> Any:
        return self.unzip(fp.name)

    @classmethod
    def unzip(cls, file_path, save_on: str = None, k=None, is_random=False):
        with ZipFile(file_path, mode='r') as myzip:
            members = sample(myzip.namelist(), k, is_random)
            if save_on:
                mkdir(save_on)
            myzip.extractall(save_on, members=members)

    def _save_fp(self, fp: Union[TextIO, BytesIO]) -> None:
        raise NotImplementedError

    @classmethod
    def load(cls, path_, load_on_memory=False, **kwargs):
        return cls(path_=path_, creation_mode=False, load_on_memory=load_on_memory, **kwargs)

    def save(self, on_new_path: Union[str, Path] = None, exist_ok=False, overwrite=False, force=False,
             **kwargs):
        with tempfile.TemporaryDirectory() as tmpdirname:
            tmp_file = Path(tmpdirname).join(self.name)
            try:
                with ZipFile(tmp_file.path, mode='w', compression=ZIP_DEFLATED) as myzip:
                    for i in self.data:
                        myzip.write(i)
            except Exception as err:
                raise IOError(f'Error writing to the file system file: {err}')
            tmp_file.cp(self.path)
            self._last_update = self.updated_at

    def parse(self, data):
        if not data:
            return []
        if isinstance(data, (str, Path)):
            data = [data]

        if not is_sequence(data):
            raise ValueError("Send list of paths invalid.")
        res = []

        for p in data:
            p = Path(p)
            if not p.exists:
                raise ValueError(f'{p.path} not Found.')
            res.append(p)
        return res


class _SrtFile(_TxtIO):
    _only_read = False
    _ext_allowed = ('.srt', '.vtt')

    @classmethod
    def is_block_start(cls, line) -> bool:
        return bool(re.match(r'\d{2}:\d{2}:\d{2}[.,]\d+\s-->\s\d{2}:\d{2}:\d{2}[.,]\d+', line))

    def parse(self, data):
        data = super().parse(data)
        blocks = []
        current_block = 0
        block = []
        block_started = False
        for i in data:
            if self.is_block_start(i):
                if block_started:
                    blocks.append(self.Block(current_block, *block))
                block_started = True
                current_block += 1
                block = []
            if block_started and not i.isdigit():
                block.append(i)
        if block:
            blocks.append(self.Block(current_block, *block))

        return blocks

    @property
    def string(self):
        return '\n'.join([str(block_item) for block in self.data for block_item in block])

    def add(self, data, line=-1):
        raise NotImplementedError

    @classmethod
    def from_url(cls, url: str, *args, **kwargs):
        # TODO: need improves
        resp = request.get(url, *args, **kwargs)
        if resp.code == 200:
            return cls(Path('./subtitle.srt'), resp.data, creation_mode=True)

    @property
    def text(self):
        return [content.strip() for block in self.data for content in block.content if content]

    class Block:
        def __init__(self, number, *content):
            self.number = number
            time_, *content = content

            self.time = time_
            start, time_separator, end = time_.split()
            self._start = start
            self._end = end
            self.content = content
            self._time_separator = time_separator

        @property
        def start(self):
            return self._start

        @property
        def end(self):
            return self._end

        @property
        def _values(self):
            return (self.number, self.time, *self.content)

        def __getitem__(self, item):
            return self._values[item]

        def __repr__(self):
            return f'{self._values}'


class FileIO:
    txt = _TxtIO
    json = _JsonIO
    mp4 = _Mp4IO
    csv = _CsvIO
    pkl = _PyObj
    zip = _ZipFileIO
    srt = _SrtFile
    _generic = _Generic

    def __new__(cls, path_, data=None, **kwargs):
        raise Exception(f"Cannot instantiate {cls.__name__}, use the methods ['create', 'load', 'load_files']")

    @classmethod
    def create(cls, path_: Union[Type[Path], str], data: Any, **kwargs) -> _FileIO:
        path_ = Path(path_)
        return cls.lookup(path_.suffix).create(path_=path_, data=data, **kwargs)

    @classmethod
    def lookup(cls, ext: str) -> Type[_FileIO]:
        ext = ext.replace('.', '')
        if hasattr(cls, ext):
            return getattr(cls, ext)
        return cls._generic

    @classmethod
    def load(cls, path_: Union[str, Path], **kwargs) -> _FileIO:
        path_ = Path(path_)
        if not path_.exists:
            raise FileNotFoundError(f"{path_.path} not found.")
        return cls.lookup(path_.suffix).load(path_=path_, **kwargs)

    @classmethod
    def load_from_clipboard(cls):
        return cls.load(Path(clipboard()))

    @classmethod
    def load_files(cls, path_, ext=None, contains_in_name: List = (), not_contains_in_name=(), take_empty=True,
                   recursive=False, ignore_dirs=()):

        ext = ext or ''
        f_paths = Path(path_).list_files(ext=ext,
                                         contains_in_name=contains_in_name,
                                         not_contains_in_name=not_contains_in_name,
                                         recursive=recursive,
                                         ignore_dirs=ignore_dirs)
        loaded = []
        for p in f_paths:
            file_ = cls.load(path_=p)
            if file_ is None:
                continue
            if take_empty is True and file_.is_empty:
                continue
            loaded.append(file_)
        return loaded


FileType = Type[_Generic]
FileTxtType = Type[_TxtIO]
FileJsonType = Type[_JsonIO]
FileCsvType = Type[_CsvIO]
FilePythonObj = Type[_PyObj]



"""

Copyright (c) 2019 The Cereja Project

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import logging
import pickle
import random
import secrets
import math
from collections import Counter, OrderedDict
from typing import Optional, Sequence, Dict, Any, List, Union, Tuple, Set

from cereja.utils._utils import is_iterable, is_sequence
from cereja.file import FileIO
from cereja.utils import invert_dict, string_to_literal
from cereja.mltools.preprocess import remove_punctuation, remove_stop_words, \
    replace_english_contractions
from cereja.utils.decorators import thread_safe_generator

__all__ = ['ConnectValues', 'Freq', 'Tokenizer', 'TfIdf', 'DataGenerator']

logger = logging.Logger(__name__)


class Freq(Counter):
    def __init__(self, data: Optional[Sequence] = ()):
        assert is_iterable(data), "Isn't iterable."
        self._elements_size = None
        super().__init__(data)

    @property
    def probability(self) -> OrderedDict:
        """
        Return ordered dict with percent
        """
        return OrderedDict(map(lambda key: (key, self.item_prob(key)), self.most_common()))

    def __getattribute__(self, item):
        if item in ('probability', 'item_prob'):
            # ensures correct probability calculation
            self._elements_size = sum(self.values())
        return super(Freq, self).__getattribute__(item)

    def sample(self, freq: Optional[int] = None, min_freq: Optional[int] = 1, max_freq: Optional[int] = None,
               max_items: int = -1,
               order='most_common'):
        """
        :param freq: only this
        :param min_freq: limit sample by frequency
        :param max_freq: limit sample by frequency
        :param max_items: limit sample size
        :param order: choice order in ('most_common', 'least_common')
        :return: sample dictionary
        """
        assert not all((freq, max_freq or min_freq != 1)), "the <freq> parameter must be used without <max_freq> and " \
                                                           "<min_freq> "
        assert order in ('most_common', 'least_common'), ValueError(
                f"Order {order} not in ('most_common', 'least_common')")
        assert isinstance(max_items, int), TypeError(f"max_items {type(max_items)} != {int}")
        max_items = len(self) if max_items == -1 else max_items
        if order == 'least_common':
            query = super(Freq, self).most_common()[:-max_items - 1:-1]
        else:
            query = super(Freq, self).most_common(max_items)
        if freq is not None:
            assert isinstance(freq, int), TypeError(f"freq {type(freq)} != {int}")
            return dict(filter(lambda item: item[1] == freq, query))

        if max_freq is not None:
            assert min_freq <= max_freq, 'min_freq > max_freq'
            query = filter(lambda item: item[1] in range(min_freq, max_freq + 1), query)
        else:
            query = filter(lambda item: item[1] in range(min_freq, item[1] + 1), query)
        return dict(query)

    def most_common(self, n: Optional[int] = None) -> Dict[Any, int]:
        return dict(super(Freq, self).most_common(n))

    def least_freq(self, max_items: int = -1):
        return self.sample(max_items=max_items, order='least_common')

    def item_prob(self, item: Any) -> float:
        if item in self:
            return self.get(item) / self._elements_size or sum(self.values())
        return 0.0

    def to_json(self, path_, probability=False, **kwargs):
        content = self.probability if probability else self.most_common()
        FileIO.create(path_=path_, data=content).save(**kwargs)


class ConnectValues(dict):
    def __init__(self, name: str = None):
        super().__init__()
        self._name = name

    @property
    def name(self):
        return self._name

    @property
    def all(self):
        for connection in self._get_connections():
            yield connection, set(self._get_connections(connection))

    def _get_connections(self, value=None):
        if value is not None:
            connections_ = filter(lambda x: value in x, self)
        else:
            connections_ = self
        for k in connections_:
            for v in k:
                yield v

    def connect(self, a, b, confidence=0.0):
        self.update({tuple(sorted((a, b))): confidence})

    def get(self, a):
        return set(self._get_connections(a))

    def confidence(self, a, b):
        return self[(tuple(sorted((a, b))))]

    def clusters(self):
        segment_matches = {}
        # I create a map containing all matches (value) of a given segment (including itself)
        for k, v in self.all:
            segment_matches[k] = v
        # cluster identifier
        cluster = 0
        clusters = {}
        while segment_matches:
            temp_matches = iter(list(segment_matches.items()))
            # remove and store all matches from this segment (set A)
            segment_id, _ = next(temp_matches)
            clusters[cluster] = segment_matches.pop(segment_id)
            while clusters[cluster].intersection(segment_matches):
                # If there is an intersection between set A and set B
                # it is understood that A is equal to B
                for k in clusters[cluster].intersection(segment_matches):
                    # for each item that is in A and B at the same time I add A and remove it from B
                    clusters[cluster].update(segment_matches.pop(k))
            cluster += 1
        return clusters


class Tokenizer:
    _n_unks = 10
    __unks = dict({f'{{{i}}}': i for i in range(_n_unks)})
    __unks.update(invert_dict(__unks))

    def __init__(self, data: Union[List[str], dict], preprocess_function=None, load_mode=False, use_unk=True):
        self._temp_unks = dict()
        self._hash = self.new_hash  # default
        self._unk_memory_max = 10000  # prevents memory leak
        self._use_unk = use_unk
        self._preprocess_function = preprocess_function
        self._warning_not_use_unk = False
        if isinstance(data, dict) and load_mode:
            logger.info("Building from file.")
            keys = data.keys()
            assert '_metadata' in keys and 'data' in keys, 'Invalid content.'
            for k, v in data['_metadata'].items():
                if k == '_preprocess_function':
                    # load function
                    if v is not None:
                        v = pickle.loads(string_to_literal(v))
                setattr(self, k, v)
            self._uniques = set(data['data'].values())
            self._index_to_item = {int(k): v for k, v in data['data'].items()}
        else:
            self._uniques = self.get_uniques(data)
            self._index_to_item = dict(enumerate(self._uniques, self._n_unks))
        self._item_to_index = invert_dict(self._index_to_item)

    @property
    def preprocess_function(self):
        return self._preprocess_function or (lambda x: x)

    @property
    def last_index(self):
        return self._n_unks + len(self._index_to_item) - 1

    def add_item(self, item):
        index_ = self.last_index + 1
        self._index_to_item[index_] = item
        self._item_to_index[item] = index_

    @property
    def unks(self):
        return self.__unks

    @property
    def unk_memory(self):
        return self._temp_unks

    @property
    def hash_default(self):
        return self._hash

    @property
    def new_hash(self):
        return secrets.token_hex(7)

    @classmethod
    def normalize(cls, data) -> List[Any]:
        if data is None:
            data = []
        if isinstance(data, str):
            return [data]
        elif isinstance(data, (int, float, bytes)):
            return [data]
        assert is_sequence(data), TypeError(f'this data type is not supported, try sending a {str}, {list} or tuple')
        return data

    def get_uniques(self, values: List[str]) -> set:
        data = self.normalize(values)
        result = []
        for v in data:
            result += v.split()
        return set(map(self.preprocess_function, result))

    def item_index(self, word: str) -> int:
        return self._item_to_index.get(word)

    def index_item(self, index: int) -> str:
        return self._index_to_item.get(index, self.unks.get(index))

    def _encode(self, data: List[str], hash__):
        if hash__ not in self._temp_unks and self._use_unk:
            if len(self._temp_unks) > self._unk_memory_max:
                self._temp_unks = {}
                logger.warning("Memory Leak Detected. Cleaned temp UNK's")
            self._temp_unks[hash__] = {}
        n = 0
        result = []
        for word in map(self.preprocess_function, data):
            index = self.item_index(word)
            if index is None:
                if not self._use_unk:
                    if not self._warning_not_use_unk:
                        logger.warning(
                                "use_unk is False. All unknown item encoded will be added to tokenizer don't forget "
                                "to save your "
                                "tokenizer after encoding.")
                        self._warning_not_use_unk = True
                    self.add_item(word)
                    index = self.last_index
                    result.append(index)
                    continue
                unk = f'{{{n % self._n_unks}}}'
                index = self.unks.get(unk)
                self._temp_unks[hash__].update({(n, f"{word}"): unk})
                n += 1
            result.append(index)
        return result

    def encode(self, data: Union[str, List[str]]) -> Union[Tuple[List[int], str], List[List[int]]]:
        """
        Encodes values in a sequence of numbers

        the hash is used to decode "unks" in the correct order.

        e.g:
            tokenizer = Tokenizer(data=['i like it', 'my name is Joab', 'hello'])
            sequences = tokenizer.encode(data=['hello my friend, how are you?', 'my name is mário'])
            # [([9, 7, 15, 10, 16, 5], 'e9bc59cb0d1564'), ([7, 13, 2, 15], '9f92140ebb0e19')]

        """
        result = []
        for sentence in self.normalize(data):
            assert isinstance(sentence, str), "send List[str] or str"
            if not self._use_unk:
                result.append(self._encode(sentence.split(), None))
                continue
            __hash = self.new_hash
            result.append((self._encode(sentence.split(), __hash), __hash))
        return result

    def decode(self, data: Union[List[int], int]):
        return [self.index_item(index) for index in self.normalize(data)]

    def to_json(self, path_: str):
        try:
            preprocess_function = str(pickle.dumps(self._preprocess_function)) if self._preprocess_function else None
        except Exception as err:
            raise Exception(f'Error on preprocess function save: {err}')
        use_unk = self._use_unk
        tokenizer_data = {'_metadata': {'_preprocess_function': preprocess_function, '_use_unk': use_unk},
                          'data':      self._index_to_item
                          }
        FileIO.create(path_, tokenizer_data).save(exist_ok=True)

    @classmethod
    def load_from_json(cls, path_: str):
        data = FileIO.load(path_)
        return cls(data.data, load_mode=True)

    def replace_unks(self, sentence: str, hash_):
        assert isinstance(sentence, str), 'expected a string.'
        try:
            _temp_unks = self._temp_unks[hash_].copy()
            if not _temp_unks:
                return sentence
            for (indx, word), unk in _temp_unks.items():  # cannot have an exception
                sentence = sentence.replace(unk, word, 1)
                self._temp_unks[hash_].pop((indx, word))
            self._temp_unks.pop(hash_)
        except KeyError as err:
            logger.error(
                    msg=f"{err}: There's something wrong open please issue "
                        f"https://github.com/jlsneto/cereja/issues/new?template=bug-report.md")
        return sentence


class TfIdf:

    def __init__(self, sentences: List[str], language: str = 'english', punctuation: str = None,
                 stop_words: List[str] = None):
        self._sentences = self.clean_sentences(sentences,
                                               language=language,
                                               punctuation=punctuation,
                                               stop_words=stop_words)
        self._corpus_bow = self._corpus_bag_of_words(self.sentences)
        self._idf = self._compute_idf(self.sentences)

    @property
    def sentences(self):
        return self._sentences

    @property
    def corpus_bow(self):
        return self._corpus_bow

    @property
    def idf(self):
        return self._idf

    @classmethod
    def sentence_bag_of_words(cls, sentence: str) -> List[str]:
        return sentence.split()

    @classmethod
    def _clean_sentence(cls, sentence: str, language: str = 'english', punctuation: str = None,
                        stop_words: List[str] = None) -> str:
        sentence = replace_english_contractions(sentence) if language == 'english' else sentence
        sentence = remove_punctuation(sentence, punctuation=punctuation)
        sentence = remove_stop_words(sentence, language=language, stop_words=stop_words)
        return sentence.lower()

    @classmethod
    def _corpus_bag_of_words(cls, sentences: List[str]):
        corpus_bow = {word.lower() for sentence in sentences for word in sentence.split()}
        print(f'{len(corpus_bow)} words on corpus')
        return corpus_bow

    @classmethod
    def _sentence_num_of_words(cls, sentence_bow: List[str]) -> Dict[str, int]:
        num_of_words = dict.fromkeys(sentence_bow, 0)
        for w in sentence_bow:
            num_of_words[w] += 1
        return num_of_words

    def _compute_idf(self, sentences: List[str]) -> Dict[str, int]:
        n = len(sentences)
        print('Computing idf...')
        idf_dict = dict.fromkeys(self.corpus_bow, 0.0)
        for sentence in sentences:
            num_of_words = self._sentence_num_of_words(self.sentence_bag_of_words(sentence))
            for word, val in num_of_words.items():
                if val > 0:
                    idf_dict[word] += 1
        for word, val in idf_dict.items():
            idf_dict[word] = math.log(n / (val + 1))
        print('idf computed!')
        return idf_dict

    @classmethod
    def clean_sentences(cls, sentences: List[str], language: str = 'english', punctuation: str = None, \
                        stop_words: List[str] = None) -> List[str]:
        return [cls._clean_sentence(sentence.lower(),
                                    language=language,
                                    punctuation=punctuation,
                                    stop_words=stop_words) for sentence in sentences]

    @classmethod
    def ordered_tf_idf(cls, sentence_tf_idf: Set[Tuple[str, float]], reverse=True):
        return sorted([word_score for word_score in sentence_tf_idf], key=lambda x: x[1], reverse=reverse)

    @classmethod
    def sentence_tf(cls, sentence_num_of_words: Dict[str, int],
                    sentence_bag_of_words: List[str]) -> Dict[str, float]:
        tf_dict = {}
        bow_count = len(sentence_bag_of_words)
        for word, count in sentence_num_of_words.items():
            tf_dict[word] = count / float(bow_count) if float(bow_count) else 0
        return tf_dict

    def sentence_tf_idf(self, sentence: str, language: str = 'english', punctuation: str = None,
                        stop_words: List[str] = None,
                        use_filter: bool = True) -> Union[Dict[str, float], Set[Tuple[str, float]]]:
        tf_idf = {}
        sentence = self._clean_sentence(sentence, language=language, punctuation=punctuation, stop_words=stop_words)
        sentence_bow = self.sentence_bag_of_words(sentence)
        sentence_now = self._sentence_num_of_words(sentence_bow)
        sentence_tf = self.sentence_tf(sentence_now, sentence_bow)
        for word, val in sentence_tf.items():
            tf_idf[word] = val * self.idf[word]
        tf_idf = set(filter(lambda x: x[1], tf_idf.items())) if use_filter else tf_idf
        return tf_idf

    def inverse_data_frequency_order(self, reverse=False):
        return sorted([(w, idf) for w, idf in self.idf.items()], key=lambda x: x[1], reverse=reverse)


class DataGenerator:
    def __init__(self, data, use_random=False):
        self._data = data
        self._use_random = use_random

    @thread_safe_generator
    def take(self, batch_size=1):
        """
        infinite loop on data

        @param batch_size: is a integer
        @return: batch of data
        """
        assert batch_size <= len(self._data), f'batch_size > data length! Send value <= {len(self._data)}'
        data = iter(self._data)
        while True:
            result = []
            if self._use_random:
                result += random.sample(self._data, k=batch_size)
            else:
                for _ in range(batch_size):
                    try:
                        result.append(next(data))
                    except StopIteration:
                        data = iter(self._data)
                        result.append(next(data))
            yield result


if __name__ == '__main__':
    tokenizer = Tokenizer(data=['i like it', 'my name is Joab', 'hello'])
    sequences = tokenizer.encode(data=['hello my friend, how are you?', 'my name is joab'])
    decoded = []
    print(sequences)
    for encoded_sequence, hash_ in sequences:
        dec = ' '.join(tokenizer.decode(encoded_sequence))
        print(dec)
        decoded.append(tokenizer.replace_unks(dec, hash_))

    print(decoded)



"""

Copyright (c) 2019 The Cereja Project

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import itertools
import warnings
from typing import List, Dict, Union, Sequence, AnyStr, Any, Iterable

from cereja import FileIO
from cereja.config.cj_types import Number
from cereja.mltools import preprocess as _preprocess
from cereja.mltools.data import Freq
from cereja.config.conf import BasicConfig
from abc import ABCMeta, abstractmethod

from cereja.system import Path

__all__ = ['LanguageData', 'Preprocessor', 'LanguageDetector']

_LANGUAGE_ISO_639_2 = {
    'eng': 'English',
    'por': 'Portuguese'
}
_LANGUAGE_ISO_639_1 = {
    'pt': 'English',
    'en': 'Portuguese'
}


def separate(text: AnyStr, sep: Union[str, Sequence[str]] = ('!', '?', '.'), between_char=False) -> str:
    warnings.warn(f"This function will be deprecated in future versions. "
                  f"preprocess.separate", DeprecationWarning, 2)
    return _preprocess.separate(text=text, sep=sep, between_char=between_char)


class LanguageConfig(BasicConfig):

    def __init__(self, name='UNK_LANG', stop_words=(),
                 punctuation='!?,.', to_lower=True, is_remove_punctuation=True, is_remove_stop_words=True,
                 is_remove_accent=False, is_destructive=False, **kwargs):
        if not isinstance(stop_words, (tuple, list)):
            raise TypeError("Stop words must be in a list or tuple")
        if not isinstance(punctuation, str):
            raise TypeError("a string is expected for punctuation")
        super().__init__(name=name, stop_words=stop_words,
                         punctuation=punctuation, to_lower=to_lower,
                         is_remove_punctuation=is_remove_punctuation,
                         is_remove_stop_words=is_remove_stop_words,
                         is_remove_accent=is_remove_accent, is_destructive=is_destructive, **kwargs)

    def _before(self, new_config: dict):
        super()._before(new_config)


class Preprocessor:
    def __init__(self, stop_words=(),
                 punctuation='!?,.', to_lower=False, is_remove_punctuation=False, is_remove_stop_words=False,
                 is_remove_accent=False, **kwargs):

        self.config = LanguageConfig(stop_words=stop_words,
                                     punctuation=punctuation, to_lower=to_lower,
                                     is_remove_punctuation=is_remove_punctuation,
                                     is_remove_stop_words=is_remove_stop_words,
                                     is_remove_accent=is_remove_accent, **kwargs)

    def __repr__(self):
        return repr(self.config)

    def _preprocess(self, sentence, is_destructive: bool):
        if is_destructive or self.config.to_lower:
            sentence = sentence.lower()
        sentence = _preprocess.remove_extra_chars(sentence)
        sentence = _preprocess.remove_non_language_elements(sentence)
        if self.config.name == 'en':
            sentence = _preprocess.replace_english_contractions(sentence)
        if is_destructive or self.config.is_remove_accent:
            sentence = _preprocess.accent_remove(sentence)
        sentence = _preprocess.separate(sentence)
        if is_destructive or self.config.is_remove_punctuation:
            sentence = _preprocess.remove_punctuation(sentence, self.config.punctuation)
        if is_destructive or self.config.is_remove_stop_words:
            sentence = ' '.join([w for w in sentence.split() if w not in self.config.stop_words])
        return _preprocess.remove_extra_chars(sentence)

    def preprocess(self, data, is_destructive=False):
        if isinstance(data, str):
            return self._preprocess(sentence=data, is_destructive=is_destructive)
        return map(lambda sentence: self._preprocess(sentence, is_destructive=is_destructive), data)


class BaseData(metaclass=ABCMeta):
    def __init__(self, data: Any):
        self._data = []
        self._prepare_data(data)

    @property
    def data(self):
        return self._data

    @abstractmethod
    def _prepare_data(self, data: Any) -> Any:
        pass


class LanguageData(BaseData):

    def __init__(self, data, stop_words=(),
                 punctuation='!?,.', to_lower=False, is_remove_punctuation=False, is_remove_stop_words=False,
                 is_remove_accent=False, **kwargs):

        self._preprocessor = Preprocessor(hook=self.re_build, stop_words=stop_words,
                                          punctuation=punctuation, to_lower=to_lower,
                                          is_remove_punctuation=is_remove_punctuation,
                                          is_remove_stop_words=is_remove_stop_words,
                                          is_remove_accent=is_remove_accent, **kwargs)

        self._phrases_freq: Freq = ...
        self._words_freq: Freq = ...
        self.__recovery_phrase_freq: Freq = ...
        self.__recovery_word_freq: Freq = ...
        super().__init__(data=data)
        self._build = True

    @property
    def config(self):
        return self._preprocessor.config

    def re_build(self):
        if hasattr(self, '_build'):
            data = self._data.copy()
            self._data.clear()
            self._phrases_freq = None
            self._words_freq = None
            self.__recovery_phrase_freq = None
            self.__recovery_word_freq = None
            self._prepare_data(data)

    def reset_freq(self):
        self._phrases_freq = self.__recovery_phrase_freq.copy()
        self._words_freq = self.__recovery_word_freq.copy()

    @property
    def vocab_size(self):
        return len(self._words_freq)

    def __contains__(self, item):
        if item in self.words_freq:
            return True
        if item in self.phrases_freq:
            return True

    def __repr__(self):
        return f'LanguageData(examples: {len(self)} - vocab_size: {self.vocab_size})'

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        return iter(self._data)

    @property
    def punctuation(self):
        return self.config.punctuation

    @property
    def words_freq(self) -> Freq:
        return self._words_freq

    @property
    def phrases_freq(self) -> Freq:
        return self._phrases_freq

    @property
    def data_prep(self):
        for phrase in self.data:
            yield self.preprocess(phrase)

    def word_freq(self, word: str):
        return self._words_freq.get(self.preprocess(word))

    @property
    def words(self):
        return set(self._words_freq)

    def preprocess(self, value: Union[str, List[str]]):
        return self._preprocessor.preprocess(value)

    def synergy(self, value: Union[Iterable[str], str]) -> Number:
        """
        Returns how important the value sent is in relation to the data set
        """
        if value is None:
            return 0.0
        if isinstance(value, str):
            value = [value]
        try:
            value = set(itertools.chain(*map(lambda val: self.preprocess(val).split(), value)))
        except AttributeError:
            raise ValueError("Inconsistent data format.")
        result = list(map(self._words_freq.item_prob, value))
        zeros = result.count(0)
        if zeros:
            return sum(result) - (zeros / len(value))
        return round(sum(result), 3)

    def save_freq(self, save_on: str, prefix='freq', ext: str = 'json', probability=False):
        ext = ext.strip('.')  # normalize
        save_on = Path(save_on)

        path_words = save_on.join(f'{prefix}_words.{ext}')
        self.words_freq.to_json(path_words, probability=probability, exist_ok=True)

        path_phrases = save_on.join(f'{prefix}_phrases.{ext}')
        self.phrases_freq.to_json(path_phrases, probability=probability, exist_ok=True)

    def _prepare_data(self, data, save_preprocessed=False):
        words = []
        phrases = []
        for phrase in data:
            prep_phrase = self.preprocess(phrase)
            if save_preprocessed:
                self._data.append(prep_phrase)
            else:
                self._data.append(phrase)

            phrases.append(prep_phrase)

            words += prep_phrase.split() if isinstance(prep_phrase, str) else [prep_phrase]

        self._words_freq = Freq(words)
        self.__recovery_word_freq = self._words_freq.copy()

        self._phrases_freq = Freq(phrases)
        self.__recovery_phrase_freq = self._phrases_freq.copy()

    def sample_words_freq(self, freq: int = None, max_items: int = -1, order='most_common'):
        return self._words_freq.sample(max_items=max_items, max_freq=freq, order=order)

    def sample_phrases_freq(self, freq: int = None, max_items: int = -1, order='most_common'):
        return self._phrases_freq.sample(max_items=max_items, max_freq=freq, order=order)


class LanguageDetector:
    __language_data: Dict[str, Union[LanguageData, dict]] = {}
    __memory = {}

    def __init__(self):
        self._is_model = False

    def load(self, filepath: str):
        self._is_model = True
        self.__language_data = FileIO.load(filepath).data
        return self

    def add(self, language: str, data_: List[str]):
        if language not in self.__language_data:
            data_ = LanguageData(data_)
            self.__language_data[language] = data_.words_freq if self._is_model else data_

    def _get_most(self, word, n_max=5):
        if word not in self.__memory:
            result = []
            for lang, obj in self.__language_data.items():
                obj = obj if self._is_model else obj.words_freq
                if word in obj:
                    result.append((lang, obj[word]))
            if result:
                self.__memory[word] = list(
                        map(lambda x: (x[0], x[1] / sum(dict(result).values())), sorted(result, key=lambda x: x[1])))
            else:
                return []
        return self.__memory[word][:n_max]

    def predict(self, sentence: str):
        result = {}
        sentence = LanguageData([sentence])
        for word in sentence.words_freq:
            if len(word) == 1:
                continue
            for lang, percent in self._get_most(word):
                percent = percent / len(sentence.words_freq)
                if lang in result:
                    result[lang] += percent
                    continue
                result[lang] = percent

        if result:
            result = sorted(result.items(), key=lambda x: x[1])
            if result[-1][1] > 0.1:
                return result[-1]
            return result

        return "UNDEFINED", 1.0

    def save(self, filepath: str):
        FileIO.create(filepath, data=self.__language_data).save(exist_ok=True)
